@@ -7,67 +7,76 @@ This software may be modified and distributed under the terms
 of the BSD license. See the LICENSE file for details.
 """
 
-from __future__ import print_function, unicode_literals, absolute_import
-
 import json
 import logging
 import os
 import re
 from contextlib import contextmanager
-from six import string_types
+from shlex import quote
 
 from .constants import DOCKERFILE_FILENAME, COMMENT_INSTRUCTION
-from .util import (b2u, extract_labels_or_envs, get_key_val_dictionary,
+from .util import (b2u, extract_key_values, get_key_val_dictionary,
                    u2b, Context, WordSplitter)
 
-try:
-    # py3
-    from shlex import quote
-except ImportError:
-    from pipes import quote
-
 
 logger = logging.getLogger(__name__)
 
 
-class Labels(dict):
+class KeyValues(dict):
+    """
+    Abstract base class for allowing direct write access to Dockerfile
+    instructions which result in a set of key value pairs.
+
+    Subclasses must override the `parser_attr` value.
+    """
+    parser_attr = None
+
+    def __init__(self, key_values, parser):
+        super(KeyValues, self).__init__(key_values)
+        self.parser = parser
+
+    def __delitem__(self, key):
+        super(KeyValues, self).__delitem__(key)
+        setattr(self.parser, self.parser_attr, dict(self))
+
+    def __setitem__(self, key, value):
+        super(KeyValues, self).__setitem__(key, value)
+        setattr(self.parser, self.parser_attr, dict(self))
+
+    def __eq__(self, other):
+        if not isinstance(other, dict):
+            return False
+        return dict(self) == other
+
+    def __hash__(self):
+        return hash(json.dumps(self, separators=(',', ':'), sort_keys=True))
+
+
+class Labels(KeyValues):
     """
     A class for allowing direct write access to Dockerfile labels, e.g.:
 
     parser.labels['label'] = 'value'
     """
-
-    def __init__(self, labels, parser):
-        super(Labels, self).__init__(labels)
-        self.parser = parser
-
-    def __delitem__(self, key):
-        super(Labels, self).__delitem__(key)
-        self.parser.labels = dict(self)
-
-    def __setitem__(self, key, value):
-        super(Labels, self).__setitem__(key, value)
-        self.parser.labels = dict(self)
+    parser_attr = 'labels'
 
 
-class Envs(dict):
+class Envs(KeyValues):
     """
     A class for allowing direct write access to Dockerfile env. vars., e.g.:
 
     parser.envs['variable_name'] = 'value'
     """
+    parser_attr = 'envs'
 
-    def __init__(self, envs, parser):
-        super(Envs, self).__init__(envs)
-        self.parser = parser
 
-    def __delitem__(self, key):
-        super(Envs, self).__delitem__(key)
-        self.parser.envs = dict(self)
+class Args(KeyValues):
+    """
+    A class for allowing direct write access to Dockerfile build args, e.g.:
 
-    def __setitem__(self, key, value):
-        super(Envs, self).__setitem__(key, value)
-        self.parser.envs = dict(self)
+    parser.args['variable_name'] = 'value'
+    """
+    parser_attr = 'args'
 
 
 class DockerfileParser(object):
@@ -75,14 +84,19 @@ class DockerfileParser(object):
                  cache_content=False,
                  env_replace=True,
                  parent_env=None,
-                 fileobj=None):
+                 fileobj=None,
+                 build_args=None):
         """
         Initialize source of Dockerfile
-        :param path: path to (directory with) Dockerfile
+        :param path: path to (directory with) Dockerfile; if not provided,
+                     and fileobj is not provided, the current working
+                     directory will be used
         :param cache_content: cache Dockerfile content inside DockerfileParser
+        :param env_replace: return content with variables replaced
         :param parent_env: python dict of inherited env vars from parent image
         :param fileobj: seekable file-like object containing Dockerfile content
                         as bytes (will be truncated on write)
+        :param build_args: python dict of build args used when building image
         """
 
         self.fileobj = fileobj
@@ -119,6 +133,13 @@ class DockerfileParser(object):
             logger.debug("Setting inherited parent image ENV vars: %s", parent_env)
             self.parent_env = parent_env
 
+        if build_args is None:
+            self.build_args = {}
+        else:
+            assert isinstance(build_args, dict)
+            logger.debug("Setting build args: %s", build_args)
+            self.build_args = build_args
+
     @contextmanager
     def _open_dockerfile(self, mode):
         if self.fileobj is not None:
@@ -141,7 +162,7 @@ class DockerfileParser(object):
 
         try:
             with self._open_dockerfile('rb') as dockerfile:
-                lines = [b2u(l) for l in dockerfile.readlines()]
+                lines = [b2u(line) for line in dockerfile.readlines()]
                 if self.cache_content:
                     self.cached_content = ''.join(lines)
                 return lines
@@ -156,11 +177,11 @@ class DockerfileParser(object):
         :param lines: list of lines to be written to Dockerfile
         """
         if self.cache_content:
-            self.cached_content = ''.join([b2u(l) for l in lines])
+            self.cached_content = ''.join(b2u(line) for line in lines)
 
         try:
             with self._open_dockerfile('wb') as dockerfile:
-                dockerfile.writelines([u2b(l) for l in lines])
+                dockerfile.writelines(u2b(line) for line in lines)
         except (IOError, OSError) as ex:
             logger.error("Couldn't write lines to dockerfile: %r", ex)
             raise
@@ -217,11 +238,11 @@ class DockerfileParser(object):
              "value": "yum -y update && yum clean all"}
         ]
         """
-        def _rstrip_backslash(l):
-            l = l.rstrip()
-            if l.endswith('\\'):
-                return l[:-1]
-            return l
+        def _rstrip_eol(text, line_continuation_char='\\'):
+            text = text.rstrip()
+            if text.endswith(line_continuation_char):
+                return text[:-1]
+            return text
 
         def _create_instruction_dict(instruction=None, value=None):
             return {
@@ -237,19 +258,39 @@ class DockerfileParser(object):
             line = re.sub(r'\n', '', line)
             return line
 
-
         instructions = []
         lineno = -1
+        line_continuation_char = '\\'
         insnre = re.compile(r'^\s*(\S+)\s+(.*)$')  # matched group is insn
         contre = re.compile(r'^.*\\\s*$')          # line continues?
         commentre = re.compile(r'^\s*#')           # line is a comment?
+        directive_possible = True
+        # escape directive regex
+        escape_directive_re = re.compile(r'^\s*#\s*escape\s*=\s*(\\|`)\s*$', re.I)
+        # syntax directive regex
+        syntax_directive_re = re.compile(r'^\s*#\s*syntax\s*=\s*(.*)\s*$', re.I)
 
         in_continuation = False
-        current_instruction = None
+        current_instruction = {}
 
         for line in self.lines:
             lineno += 1
 
+            if directive_possible:
+                # once support for python versions before 3.8 is dropped use walrus operator
+                if escape_directive_re.match(line):
+                    # Do the matching twice if there is a directive to avoid doing the matching
+                    # for other lines
+                    match = escape_directive_re.match(line)
+                    line_continuation_char = match.group(1)
+                    contre = re.compile(r'^.*' + re.escape(match.group(1)) + r'\s*$')
+                elif syntax_directive_re.match(line):
+                    # Currently no information for the syntax directive is stored it is still
+                    # necessary to detect escape directives after a syntax directive
+                    pass
+                else:
+                    directive_possible = False
+
             # It is necessary to keep instructions and comment parsing separate,
             # as a multi-line instruction can be interjected with comments.
             if commentre.match(line):
@@ -266,21 +307,20 @@ class DockerfileParser(object):
                         continue
                     current_instruction = _create_instruction_dict(
                         instruction=m.groups()[0].upper(),
-                        value=_rstrip_backslash(m.groups()[1])
+                        value=_rstrip_eol(m.groups()[1], line_continuation_char)
                     )
                 else:
                     current_instruction['content'] += line
                     current_instruction['endline'] = lineno
 
-                    # pylint: disable=unsupported-assignment-operation
                     if current_instruction['value']:
-                        current_instruction['value'] += _rstrip_backslash(line)
+                        current_instruction['value'] += _rstrip_eol(line, line_continuation_char)
                     else:
-                        current_instruction['value'] = _rstrip_backslash(line.lstrip())
-                    # pylint: enable=unsupported-assignment-operation
+                        current_instruction['value'] = _rstrip_eol(line.lstrip(),
+                                                                   line_continuation_char)
 
                 in_continuation = contre.match(line)
-                if not in_continuation and current_instruction is not None:
+                if not in_continuation and current_instruction:
                     instructions.append(current_instruction)
 
         return instructions
@@ -298,13 +338,26 @@ class DockerfileParser(object):
         """
         :return: list of parent images -- one image per each stage's FROM instruction
         """
+        in_stage = False
+        top_args = {}
         parents = []
         for instr in self.structure:
-            if instr['instruction'] != 'FROM':
-                continue
-            image, _ = image_from(instr['value'])
-            if image is not None:
-                parents.append(image)
+            if instr['instruction'] == 'ARG':
+                if not in_stage:
+                    key_val_list = extract_key_values(
+                        env_replace=False,
+                        args={}, envs={},
+                        instruction_value=instr['value'])
+                    for key, value in key_val_list:
+                        if key in self.build_args:
+                            value = self.build_args[key]
+                        top_args[key] = value
+            elif instr['instruction'] == 'FROM':
+                in_stage = True
+                image, _ = image_from(instr['value'])
+                if image is not None:
+                    image = WordSplitter(image, args=top_args).dequote()
+                    parents.append(image)
         return parents
 
     @parent_images.setter
@@ -324,7 +377,7 @@ class DockerfileParser(object):
                 continue
 
             old_image, stage = image_from(instr['value'])
-            if not old_image:
+            if old_image is None:
                 continue  # broken FROM, fixing would just confuse things
             if not parents:
                 raise RuntimeError("not enough parents to match build stages")
@@ -360,7 +413,14 @@ class DockerfileParser(object):
         """
         change image of final stage FROM instruction
         """
-        images = self.parent_images or [None]
+        images = []
+        for instr in self.structure:
+            if instr['instruction'] == 'FROM':
+                image, _ = image_from(instr['value'])
+                if image is not None:
+                    images.append(image)
+        if not images:
+            raise RuntimeError('No stage defined to set base image on')
         images[-1] = new_image
         self.parent_images = images
 
@@ -414,30 +474,55 @@ class DockerfileParser(object):
         """
         return self._instruction_getter('ENV', env_replace=self.env_replace)
 
+    @property
+    def args(self):
+        """
+        ARGs from Dockerfile
+        :return: dictionary of arg_var_name:value (value might be '')
+        """
+        return self._instruction_getter('ARG', env_replace=self.env_replace)
+
     def _instruction_getter(self, name, env_replace):
         """
-        Get LABEL or ENV instructions with environment replacement
+        Get LABEL or ENV or ARG instructions with environment replacement
 
-        :param name: e.g. 'LABEL' or 'ENV'
+        :param name: e.g. 'LABEL' or 'ENV' or 'ARG'
         :param env_replace: bool, whether to perform ENV substitution
         :return: Labels instance or Envs instance
         """
-        if name != 'LABEL' and name != 'ENV':
-            raise ValueError("Unsupported instruction '%s'" % name)
+        if name not in ('LABEL', 'ENV', 'ARG'):
+            raise ValueError("Unsupported instruction '{0}'".format(name))
+        in_stage = False
+        top_args = {}
         instructions = {}
+        args = {}
         envs = {}
 
         for instruction_desc in self.structure:
             this_instruction = instruction_desc['instruction']
             if this_instruction == 'FROM':
+                in_stage = True
                 instructions.clear()
+                args = {}
                 envs = self.parent_env.copy()
-            elif this_instruction in (name, 'ENV'):
+            elif this_instruction in (name, 'ENV', 'ARG'):
                 logger.debug("%s value: %r", name.lower(), instruction_desc['value'])
-                key_val_list = extract_labels_or_envs(env_replace=env_replace,
-                                                      envs=envs,
-                                                      instruction_value=instruction_desc['value'])
+                key_val_list = extract_key_values(
+                    env_replace=this_instruction != 'ARG' and env_replace,
+                    args=args, envs=envs,
+                    instruction_value=instruction_desc['value'])
                 for key, value in key_val_list:
+                    if this_instruction == 'ARG':
+                        if in_stage:
+                            if key in top_args:
+                                value = top_args[key]
+                            elif key in self.build_args:
+                                value = self.build_args[key]
+                            args[key] = value
+                        else:
+                            if key in self.build_args:
+                                value = self.build_args[key]
+                            top_args[key] = value
                     if this_instruction == name:
                         instructions[key] = value
                         logger.debug("new %s %r=%r", name.lower(), key, value)
@@ -445,7 +530,12 @@ class DockerfileParser(object):
                         envs[key] = value
 
         logger.debug("instructions: %r", instructions)
-        return Labels(instructions, self) if name == 'LABEL' else Envs(instructions, self)
+        if name == 'LABEL':
+            return Labels(instructions, self)
+        elif name == 'ENV':
+            return Envs(instructions, self)
+        else:
+            return Args(instructions, self)
 
     @labels.setter
     def labels(self, labels):
@@ -463,6 +553,14 @@ class DockerfileParser(object):
         """
         self._instructions_setter('ENV', envs)
 
+    @args.setter
+    def args(self, args):
+        """
+        Setter for ARG instruction, i.e. sets ARGs per input param.
+        :param args: dictionary of arg names & values to be set
+        """
+        self._instructions_setter('ARG', args)
+
     def _instructions_setter(self, name, instructions):
         if not isinstance(instructions, dict):
             raise TypeError('instructions needs to be a dictionary {name: value}')
@@ -471,6 +569,10 @@ class DockerfileParser(object):
             existing = self.labels
         elif name == 'ENV':
             existing = self.envs
+        elif name == 'ARG':
+            existing = self.args
+        else:
+            raise ValueError("Unexpected instruction '%s'" % name)
 
         logger.debug("setting %s instructions: %r", name, instructions)
 
@@ -496,6 +598,9 @@ class DockerfileParser(object):
     def _modify_instruction_env(self, env_var_key, env_var_value):
         self._modify_instruction_label_env('ENV', env_var_key, env_var_value)
 
+    def _modify_instruction_arg(self, arg_key, arg_value):
+        self._modify_instruction_label_env('ARG', arg_key, arg_value)
+
     def _modify_instruction_label_env(self, instruction, instr_key, instr_value):
         """
         set <INSTRUCTION> instr_key to instr_value
@@ -507,6 +612,8 @@ class DockerfileParser(object):
             instructions = self.labels
         elif instruction == 'ENV':
             instructions = self.envs
+        elif instruction == 'ARG':
+            instructions = self.args
         else:
             raise ValueError("Unknown instruction '%s'" % instruction)
 
@@ -526,8 +633,8 @@ class DockerfileParser(object):
         for candidate in candidates:
             words = list(WordSplitter(candidate['value']).split(dequote=False))
 
-            # LABEL/ENV syntax is one of two types:
-            if '=' not in words[0]:  # LABEL/ENV name value
+            # LABEL/ENV/ARG syntax is one of two types:
+            if '=' not in words[0]:  # LABEL/ENV/ARG name value
                 # Remove quotes from key name and see if it's the one
                 # we're looking for.
                 if WordSplitter(words[0]).dequote() == instr_key:
@@ -544,7 +651,7 @@ class DockerfileParser(object):
                     startline = candidate['startline']
                     endline = candidate['endline']
                     break
-            else:  # LABEL/ENV "name"="value"
+            else:  # LABEL/ENV/ARG "name"="value"
                 for index, token in enumerate(words):
                     key, _ = token.split("=", 1)
                     if WordSplitter(key).dequote() == instr_key:
@@ -589,6 +696,9 @@ class DockerfileParser(object):
         if instruction == 'ENV' and value:
             self._modify_instruction_env(value, None)
             return
+        if instruction == 'ARG' and value:
+            self._modify_instruction_arg(value, None)
+            return
 
         lines = self.lines
         deleted = False
@@ -606,7 +716,7 @@ class DockerfileParser(object):
         :param instruction: instruction name to be added
         :param value: instruction value
         """
-        if (instruction == 'LABEL' or instruction == 'ENV') and len(value) == 2:
+        if instruction in ('LABEL', 'ENV', 'ARG') and len(value) == 2:
             new_line = instruction + ' ' + '='.join(map(quote, value)) + '\n'
         else:
             new_line = '{0} {1}\n'.format(instruction, value)
@@ -652,7 +762,8 @@ class DockerfileParser(object):
         for stage in range(len(froms)-2, -1, -1):  # e.g. 0 for single or 2, 1, 0 for 3 stages
             start, finish = froms[stage], froms[stage+1]
             linenum = start['endline'] + 1 if at_start else finish['startline']
-            if skip_scratch and froms[stage]['value'] == 'scratch':
+            image, _ = image_from(froms[stage].get('value') or '')
+            if skip_scratch and image == 'scratch':
                 continue
             df_lines[linenum:linenum] = lines
 
@@ -686,7 +797,7 @@ class DockerfileParser(object):
                 anchor = anchor['endline']
             else:
                 anchor = anchor['startline']
-        elif isinstance(anchor, string_types):  # line contents
+        elif isinstance(anchor, str):  # line contents
             matches = [index for index, text in enumerate(df_lines) if text == anchor]
             if not matches:
                 raise RuntimeError("Cannot find line in the build file:\n" + anchor)
@@ -708,23 +819,42 @@ class DockerfileParser(object):
     def context_structure(self):
         """
         :return: list of Context objects
-            (Contains info about labels and environment variables for each line.)
+            (Contains info about build arguments, labels, and environment variables for each line.)
         """
+        in_stage = False
+        top_args = {}
         instructions = []
         last_context = Context()
         for instr in self.structure:
             instruction_type = instr['instruction']
             if instruction_type == "FROM":  # reset per stage
-                last_context = Context()
+                in_stage = True
+                last_context = Context(envs=dict(self.parent_env))
 
-            context = Context(envs=dict(last_context.envs),
+            context = Context(args=dict(last_context.args),
+                              envs=dict(last_context.envs),
                               labels=dict(last_context.labels))
 
-            if instruction_type in ["ENV", "LABEL"]:
-                val = get_key_val_dictionary(instruction_value=instr['value'],
-                                             env_replace=self.env_replace,
-                                             envs=last_context.envs)
-                context.set_line_value(context_type=instruction_type, value=val)
+            if instruction_type in ('ARG', 'ENV', 'LABEL'):
+                values = get_key_val_dictionary(
+                    instruction_value=instr['value'],
+                    env_replace=instruction_type != 'ARG' and self.env_replace,
+                    args=last_context.args,
+                    envs=last_context.envs)
+                if instruction_type == 'ARG' and self.env_replace:
+                    if in_stage:
+                        for key in list(values.keys()):
+                            if key in top_args:
+                                values[key] = top_args[key]
+                            elif key in self.build_args:
+                                values[key] = self.build_args[key]
+                    else:
+                        for key, value in list(values.items()):
+                            if key in self.build_args:
+                                value = self.build_args[key]
+                            top_args[key] = value
+                            values[key] = value
+                context.set_line_value(context_type=instruction_type, value=values)
 
             instructions.append(context)
             last_context = context
@@ -736,10 +866,12 @@ def image_from(from_value):
     :param from_value: string like "image:tag" or "image:tag AS name"
     :return: tuple of the image and stage name, e.g. ("image:tag", None)
     """
-    regex = re.compile(r"""(?xi)     # readable, case-insensitive regex
-        \s*                          # ignore leading whitespace
-        (?P<image> \S+ )             # image and optional tag
-        (?:                          # optional "AS name" clause for stage
+    regex = re.compile(r"""(?xi)        # readable, case-insensitive regex
+        \s*                             # ignore leading whitespace
+        (?P<platform> --platform=\S+)?  # optional platform parameter
+        \s*                             # more whitespaces
+        (?P<image> \S+ )                # image and optional tag
+        (?:                             # optional "AS name" clause for stage
             \s+ AS \s+
             (?P<name> \S+ )
         )?
