@@ -7,10 +7,10 @@ This software may be modified and distributed under the terms
 of the BSD license. See the LICENSE file for details.
 """
 
-from __future__ import unicode_literals, absolute_import
+import io
 
 import pytest
-import six
+
 
 from dockerfile_parse import DockerfileParser
 
@@ -29,14 +29,14 @@ def dfparser(tmpdir, request):
 
     use_fileobj, cache_content = request.param
     if use_fileobj:
-        fileobj = six.BytesIO()
+        fileobj = io.BytesIO()
         return DockerfileParser(fileobj=fileobj, cache_content=cache_content)
     else:
         tmpdir_path = str(tmpdir.realpath())
         return DockerfileParser(path=tmpdir_path, cache_content=cache_content)
 
 
-@pytest.fixture(params=['LABEL', 'ENV'])
+@pytest.fixture(params=['LABEL', 'ENV', 'ARG'])
 def instruction(request):
     """
     Parametrized fixture which enables to run a test once for each instruction in params
